import logging
import time


class GSDBSLogger(logging.Handler):
    '''
    Customized logging handler that puts logs to the database.
    pymssql required
    '''

    def __init__(self, sql_conn, sql_cursor, db_tbl_log):
        logging.Handler.__init__(self)
        self.sql_cursor = sql_cursor
        self.sql_conn = sql_conn
        self.db_tbl_log = db_tbl_log
        self.cnt = 0

    def emit(self, record, jobinfo):
        self.cnt += 1
        tm = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(record.created))
        self.log_msg = record.msg
        self.log_msg = self.log_msg.strip()
        self.log_msg = self.log_msg.replace('\'', '\'\'')
        # Make the SQL insert
        gql = f"""
           mutation{{
            addDTable(
                dtablename:"gscnodelog",
                superDTable:[DTABLE],
                sriBuildInfo:"${{jobid}}-${{groupid}}-${{computingstep}}-${{cnode}}-${{cnt}}",
                dataLinks:[ {{ alias:"jobid",locale:DE,superPropertyURI:DYNAMIC_DATALINK,DataType:STRING}},
                            {{ alias:"groupid",locale:DE,superPropertyURI:DYNAMIC_DATALINK,DataType:STRING}},
                            {{ alias:"computingstep",locale:DE,superPropertyURI:DYNAMIC_DATALINK,DataType:STRING}},
                            {{ alias:"cnode",locale:DE,superPropertyURI:DYNAMIC_DATALINK,DataType:STRING}},
                            {{ alias:"cnt",locale:DE,superPropertyURI:DYNAMIC_DATALINK,DataType:STRING}},
                            {{ alias:"level",locale:DE,superPropertyURI:DYNAMIC_DATALINK,DataType:STRING}},
                            {{ alias:"levelname",locale:DE,superPropertyURI:DYNAMIC_DATALINK,DataType:STRING}},
                            {{ alias:"log",locale:DE,superPropertyURI:DYNAMIC_DATALINK,DataType:STRING}},
                            {{ alias:"createdat",locale:DE,superPropertyURI:DYNAMIC_DATALINK,DataType:STRING}}],
                data:[["jobid","groupid","computingstep","cnode","cnt","level","levelname","log","createdat"],
                      ["${jobinfo["jobid"]}"  ,"${""}"    ,"${""}"          ,"${""}"  ,"${""}","${""}"  ,"${""}"      ,"${""}","${""}"]]
            )
        }}
        """

#
logging.getLogger('').addHandler(logdb)
#
# logdb = GSDBSLogger(log_conn, log_cursor, db_tbl_log)



import base64
import logging

import blosc
import numpy as np

from gsdbs.dbclient import GSDBS


def gsNpCompressBase64WithShape(frame):
    bytes_array = frame.tostring()
    zframe = blosc.compress(bytes_array)
    b64frame = base64.b64encode((zframe))
    return b64frame.decode('utf-8'), frame.shape[0], frame.shape[1], frame.shape[2]


def gsNpCompressWithShape(frame):
    bytes_array = frame.tostring()
    zframe = blosc.compress(bytes_array)
    return zframe, frame.shape[0], frame.shape[1], frame.shape[2]


def gsNumpyDecompressBase64(b64frame, w, h, d):
    zframe = base64.b64decode(b64frame.encode('utf-8'))
    frame = np.frombuffer(blosc.decompress(zframe, True), dtype=np.uint8)
    frame.shape = (int(w), int(h), int(d))
    return frame


def gsNumpyDecompress(zframe, w, h, d):
    frame = np.frombuffer(blosc.decompress(zframe, True), dtype=np.uint8)
    frame.shape = (int(w), int(h), int(d))
    return frame


def gsNpToBytesWithShape(frame):
    return frame.tobytes(), frame.shape[0], frame.shape[1], frame.shape[2]


def gsBytesToNPArray(frame):
    nparray = np.frombuffer(frame[0], dtype=np.uint8)
    nparray.shape = (int(frame[1]), int(frame[2]), int(frame[3]))
    return nparray


def handleparametermissing(request, *args):
    for parameter in args:
        if request.get(parameter) is None:
            raise ValueError('Missing paramter: ' + parameter)


class LogDBHandler(logging.Handler):
    '''
    Customized logging handler that puts logs to the database.
    pymssql required
    '''

    def __init__(self, gsdbs: GSDBS):
        logging.Handler.__init__(self)
        self.gsdbs = gsdbs

    def emit(self, record):
        # Set current time
        # Clear the log message so it can be put to db via sql (escape quotes)
        self.log_msg = record.msg
        self.log_msg = self.log_msg.strip()
        self.log_msg = self.log_msg.replace('\'', '\'\'')
        logrequest = f"""
            mutation{{
            addDTable(
                dtablename:"gslog",
                superDTable:[DTABLE],
                sriBuildInfo:"${{jobid}}",
                dataLinks:[{{ alias:"jobid",locale:DE,superPropertyURI:DYNAMIC_DATALINK,DataType:STRING}},
                {{ alias:"groupid",locale:DE,superPropertyURI:DYNAMIC_DATALINK,DataType:STRING}},
                {{ alias:"computingstep",locale:DE,superPropertyURI:DYNAMIC_DATALINK,DataType:STRING}},
                {{ alias:"cnode",locale:DE,superPropertyURI:DYNAMIC_DATALINK,DataType:STRING}},
                {{ alias:"level",locale:DE,superPropertyURI:DYNAMIC_DATALINK,DataType:STRING}},
                {{ alias:"message",locale:DE,superPropertyURI:DYNAMIC_DATALINK,DataType:STRING}}],
                data:[
                    ["jobid","groupid","computingstep","cnode","level","message"]
                    ["{self.gsdbs.credentials["cnode"]}","","","{self.gsdbs.credentials["cnode"]}","{str(record.levelname)}","{str(self.log_msg)}"]
                ]
            )
        }}
        """
        self.gsdbs.executeStatementSync(logrequest)



import ast
import asyncio
import ctypes
import http.client as httplib
import os
import random
import sys
import threading
import time
from datetime import datetime

import websocket
import string
from threading import Thread
import json
import requests
from websocket import WebSocketConnectionClosedException

from gsdbs.dbclient import GSDBS
from twisted.internet import task, reactor
from multiprocessing.dummy import Pool
import logging
import os.path
from multiprocessing import Process, Queue


class MissionControl(Thread):

    def __init__(self, cnode, prefix, gsdbs: GSDBS, execute, ondestruct, restartCNode, heartbeatintervall=10):
        self._mandantname = ""
        self.counter = 0
        self.cnode = cnode
        self._gsdbs = gsdbs
        self._prefix = prefix
        self.execute = execute
        self.ondestruct = ondestruct
        self.restartCNode = restartCNode
        self.heartbeatintervall = heartbeatintervall * 1000
        self._pool = Pool(self._gsdbs.credentials["poolsize"])
        self._logger = logging.getLogger(__name__)
        self.joblist = {}
        Thread.__init__(self)
        self.threadid = threading.current_thread().ident
        self.connect()

    def connect(self):
        self.start()

    def disconnect(self):
        pass

    def run(self):
        self._r1 = str(random.randint(0, 1000))
        self._conn_id = self.random_str(8)
        websocket.enableTrace(False)

        if len(self._gsdbs.cookiejar) == 0:
            self._logger.error("no session available restarting")
            exit(-1)

        self._ws = websocket.WebSocketApp(
            ("ws://" if "localhost" in self._gsdbs.credentials["wshost"] else "wss://")
            + self._gsdbs.credentials["wshost"] + ":" + str(self._gsdbs.credentials["wsport"]) +
            self._prefix +
            "/" +
            self._r1 +
            "/" +
            self._conn_id +
            "/websocket?gssession=" +
            self._gsdbs.cookiejar.get("session") + "." + self._gsdbs.cookiejar.get("signature"),
            on_message=self.on_message,
            on_open=self.on_open,
            on_error=self.on_error,
            on_close=self.on_close)

        self._ws.run_forever()

    def on_message(self, ws, message):
        self._pool.apply_async(self.processMessage, args=[message], callback=self.on_success,
                               error_callback=self.on_errorPost)

    def on_errorPost(self, error):
        self._logger.exception('message failed :' + error)

    def processMessage(self, message):
        # sleep(0.05)

        if message == 'a["\\n"]':
            self._logger.debug("beat:" + datetime.now().strftime("%H:%M:%S"))
            self._ws.send('["\\n"]')
            self.counter = self.counter + 1
            # if self.counter % self.heartbeatintervall == 0:
            self.ETHome()
            # self.counter = 0

        if message == "o":
            pass
        if message.startswith("a"):
            if "{" in message:
                ackid = list(filter(lambda c: c.startswith('message-id'), message.split("\\n")))[0].split(":")[1]
                subscription = list(filter(lambda c: c.startswith('subscription'), message.split("\\n")))[0].split(":")[
                    1]
                ack = f'ACK\\nid:{ackid}\\nsubscription:{subscription}\\n\\n\\u0000'
                self._ws.send("[\"" + ack + "\"]")
                try:
                    mssgbdy = json.loads(
                        message[message.find("{"):message.find("\\u0000")].replace("\\\"", "\"").replace("\\n", " ")
                        .replace("\\r", " ").replace("\\t", " "))

                    if "gscnoderestart" in mssgbdy:
                        self.restartCNode(self.threadid)

                    if "gscnodestatus" in mssgbdy:
                        self.reportJobStatus(mssgbdy)
                    else:
                        self._logger.info("Starting job")
                        if len(self.joblist) < self._gsdbs.credentials["concurrentjobcount"]:
                            self.joblist[mssgbdy["jobid"]] = mssgbdy
                            self.execute(self._gsdbs, mssgbdy, self.onNext)
                        else:
                            self._logger.warn("max concurrent job count reached")
                            return

                    if "isCommand" not in mssgbdy and "gscnodestatus" not in mssgbdy and "gscnoderestart" not in mssgbdy:
                        if self.str2bool(mssgbdy["isComputingStep"]) and mssgbdy["computingstep"] != '':
                            self.markJobAsDone(mssgbdy["jobid"], mssgbdy["groupid"], mssgbdy["computingstep"])

                except Exception as e:
                    self._logger.exception(e)
                    if "mssgbdy" in locals():
                        if "gscnoderestart" not in mssgbdy:
                            self.markJobAsFailed(mssgbdy["jobid"])
                    else:
                        self._logger.exception("JobFailed " + str(e))
                    if self.ondestruct is not None:
                        self.ondestruct()

        if message == "start_streaming":
            self.execute(self._gsdbs, {"start_streaming": True}, self.onNext)

        else:
            pass

    def str2bool(self, v):
        if isinstance(v, (bool)):
            return v
        else:
            return v.lower() in ("yes", "true", "t", "1")

    def call_api(self, send):
        x = 16388
        res = [send[y - x:y] for y in range(x, len(send) + x, x)]
        for x in res:
            self._ws.send("[\"" + x + "\"]")

    def on_success(self, r):
        self._logger.debug('Job succeed')

    def on_errorPost(self, error):
        self._logger.error('job failed ' + str(error))

    def onNext(self, target, json1, data):
        try:
            data["mandantname"] = self._mandantname

            url = "http://" + self._gsdbs.credentials["wshost"] + ":" + str(
                self._gsdbs.credentials["wsport"]) + "/missioncontrol/onnext"
            json2 = {"jobid": json1["jobid"],
                     "groupid": json1["groupid"],
                     "accesstoken": self._gsdbs.cookiejar.get("session") + "." + self._gsdbs.cookiejar.get("signature"),
                     "computingstep": json1["computingstep"],
                     "target": target,
                     "cnode": self.cnode,
                     "jobkonfigname": json1["jobconfigname"],
                     "data": data}

            datastring = json.dumps(json2).replace("\"", "\'")
            self._logger.debug(datastring)
            transactionid = 'tx-' + self.random_str(8)
            send = 'SEND\\n' \
                   'destination:/queue/onnext\\n' \
                   'durable:false\\n' \
                   'exclusive:false\\n' \
                   'auto-delete:false\\n\\n' \
                   + datastring + '\\u0000'
            self.call_api(send)

        except Exception as e:
            self._logger.exception(e)
            return e

    def utf8len(self, s):

        encoded_string = s.encode('utf-8')
        byte_array = bytearray(s)

        return len(s.encode('utf-8'))

    def reportJobStatus(self, mssgbdy):

        targetJobName = mssgbdy["jobconfigname"]
        jobconfigFound = False

        for jobinfo in self.joblist:
            testSameJobkonfigname = self.joblist[jobinfo]["jobconfigname"] == targetJobName
            if testSameJobkonfigname:
                jobconfigFound = True
                break

        if not jobconfigFound:
            reportAnswer = {'channel': mssgbdy["streamkey"],
                            'pipelinetemplateName': mssgbdy["pipelinetemplateName"],
                            'jobconfigname': targetJobName,
                            'envmap': mssgbdy
                            }

            resp = requests.post(self._gsdbs.credentials["baseurl"] + "/missioncontrol/channelpipelines/rva/report",
                                 json=reportAnswer,
                                 cookies=self._gsdbs.cookiejar)
            resp.raise_for_status()

    def on_error(self, ws, error):
        self._logger.exception(error)
        if not (type(error) is (
                ConnectionRefusedError or ConnectionAbortedError or ConnectionResetError or WebSocketConnectionClosedException)):
            statuscode = getattr(error, 'status_code', 'default value')
            if statuscode == "status_code":
                self._gsdbs.getTokenFromApi()

    def on_close(self, ws, close_status_code, close_msg):
        if close_status_code == 401:
            self._gsdbs.getTokenFromApi()

        self._logger.exception(
            "### closed: Code->" + str(close_status_code) + "Message:" + str(close_msg) + ":" + datetime.now().strftime(
                "%H:%M:%S") + "###")

    def on_open(self, ws):
        connect = '\"CONNECT\\naccept-version:1.2\\nheart-beat:' + str(self.heartbeatintervall) + ',' + str(
            self.heartbeatintervall) + '\\nprefetch-count:1\\nauto-delete:true\\nexclusive:true\\nx-single-active-consumer:true\\n\\n\\u0000\"'
        self._ws.send("[" + connect + "]")
        time.sleep(1)
        sub = f'\"SUBSCRIBE\\nid:{self.random_str(4)}\\ndestination:/queue/{self.getQueue()}\\nack:client-individual\\n\\n\\u0000\"'
        self._ws.send("[" + sub + "]")
        self.ETHome()
        self._logger.info("open:" + datetime.now().strftime("%H:%M:%S"))
        if "camera" in self._gsdbs.credentials and self._gsdbs.credentials["camera"]:
            self.on_message(None, "start_streaming")

    def ETHome(self):
        resp = requests.get(self._gsdbs.credentials["baseurl"] + "missioncontrol/register",
                            cookies=self._gsdbs.cookiejar)
        resp.raise_for_status()

        if len(self.joblist) == 0:
            resp = requests.get(self._gsdbs.credentials["baseurl"] + "missioncontrol/channelpipelines/rva/heartbeat",
                                cookies=self._gsdbs.cookiejar)

        # self._gsdbs.executeStatementSync(f"""
        #                                     mutation{{
        #                                       addDTable(
        #                                         dtablename:"cnode",
        #                                         superDTable:DTABLE,
        #                                         sriBuildInfo:"${{cnode}}",
        #                                         dataLinks:[
        #                                           {{alias:"cnode",locale:DE,superPropertyURI:DYNAMIC_DATALINK,DataType:STRING}}
        #                                           {{alias:"desciption",locale:DE,superPropertyURI:DYNAMIC_DATALINK,DataType:STRING}}
        #                                         ]
        #                                         data:[
        #                                           ["cnode","desciption"],
        #                                           ["{self.cnode}","desciption"]
        #                                         ]
        #                                       )
        #                                     }}
        #                                     """)

        # if "camera" in self._gsdbs.credentials and self._gsdbs.credentials["camera"]:
        #
        #     self._gsdbs.executeStatementSync(f"""
        #                         mutation{{
        #                               addDTable(
        #                                 dtablename:"gscamera",
        #                                 superDTable:DTABLE,
        #                                 sriBuildInfo:"${{streamkey}}",
        #                                 dataLinks:[
        #                                   {{alias:"streamkey",locale:DE,superPropertyURI:DYNAMIC_DATALINK,DataType:STRING}}
        #                                   {{alias:"name",locale:DE,superPropertyURI:DYNAMIC_DATALINK,DataType:STRING}}
        #                                   {{alias:"humidity",locale:DE,superPropertyURI:DYNAMIC_DATALINK,DataType:STRING}}
        #                                   {{alias:"temperatur",locale:DE,superPropertyURI:DYNAMIC_DATALINK,DataType:STRING}}
        #                                   {{alias:"hres",locale:DE,superPropertyURI:DYNAMIC_DATALINK,DataType:STRING}}
        #                                   {{alias:"vres",locale:DE,superPropertyURI:DYNAMIC_DATALINK,DataType:STRING}}
        #                                   {{alias:"fps",locale:DE,superPropertyURI:DYNAMIC_DATALINK,DataType:STRING}}
        #                                   {{alias:"ptz",locale:DE,superPropertyURI:DYNAMIC_DATALINK,DataType:STRING}}
        #                                 ],
        #                                 data:[
        #                                     ["streamkey","name","humidity","temperatur","hres","vres","fps","ptz"]
        #                                      ["{self._gsdbs.credentials["cnode"]}","{self._gsdbs.credentials["cnode"]}","0","0","{self._gsdbs.credentials["hres"]}","{self._gsdbs.credentials["vres"]}","{self._gsdbs.credentials["framerate"]}","{self._gsdbs.credentials["ptz"]}"]
        #
        #                                 ]
        #                               )
        #                     }}
        #                     """)

    def getQueue(self):
        try:
            resp = requests.get(self._gsdbs.credentials["baseurl"] + self._gsdbs.credentials['userInfo'],
                                cookies=self._gsdbs.cookiejar)
            resp.raise_for_status()
            userinfo = resp.json()
            self._mandantname = userinfo["mandant"]["mandantName"]
            return userinfo["mandant"]["mandantName"] + "-" + self.cnode  # + "-" + self._conn_id
        except Exception as e:
            self._logger.exception("cant get quere" + str(e))
            return ""

    def random_str(self, length):
        letters = string.ascii_lowercase + string.digits
        return ''.join(random.choice(letters) for c in range(length))

    def markJobAsDone(self, jobid, groupid, computingstep):
        self.joblist.pop(jobid)
        json2 = {"jobid": jobid, "groupid": groupid,
                 "accesstoken": self._gsdbs.cookiejar.get("session") + "." + self._gsdbs.cookiejar.get("signature"),
                 "computingstep": computingstep, "cnode": self.cnode}
        datastring = json.dumps(json2).replace("\"", "\\\"")

        send = 'SEND\\ndestination:/queue/onnotify\\ndurable:false\\nexclusive:false\\nauto-delete:false\\n\\n' + datastring + '\\u0000'
        self.call_api(send)
        if self.ondestruct is not None:
            self.ondestruct(self._gsdbs)
        self._logger.info("job done")

    def markJobAsFailed(self, jobid):
        self.joblist.pop(jobid)
        self._gsdbs.executeStatement(f"""
                    mutation{{
                            updateDTable(
                      dtablename:gsasyncjob,
                       where: [
                          {{connective: BLANK, column: gsasyncjob_jobid, operator: EQUAL, value: "{jobid}"}}
                            {{connective: AND, column: gsasyncjob_cnode, operator: EQUAL, value: "{self.cnode}"}}
                      ],
                      updatelist:[
                        {{datalink:gsasyncjob_jobstatus,value:"failed"}}
                      ]
                    )
                    }}
                """)


class MissionControlClient:

    def __init__(self, execute,
                 gsdbspath=os.path.dirname(os.path.realpath(__file__)),
                 configstr="",
                 logginglevel=logging.WARNING,
                 initmethod=None,
                 ondestruct=None):
        self.threadident = None
        self.execute = execute
        self.gsdbs = GSDBS(gsdbspath, configstr=configstr)
        self.cnode = self.gsdbs.credentials["cnode"]
        self.client = None
        self.logginglevel = self.gsdbs.credentials["logginglevel"]
        self.initMethod = initmethod
        self.reactor = reactor
        self.ondestruct = ondestruct
        self.restartPlanned = False
        self.logger = logging.getLogger(__name__)
        self.init()

    def createClient(self):
        try:
            self.client = MissionControl(self.cnode,
                                         '/gs-guide-websocket',
                                         self.gsdbs,
                                         self.execute,
                                         self.ondestruct,
                                         self.restartCNode,
                                         60)
        except Exception as e:
            self.logger.exception("critical cant get connection" + str(e))
            os._exit(0)

    def restartCNode(self, threadid):
        self.reactor.callFromThread(self.reactor.stop)
        # self.terminate()

    def checkThreadRunning(self):
        if not self.restartPlanned:
            if self.client is None or not self.client.is_alive():
                reactor.callFromThread(self.createClient)

    def terminate(self):
        # os.system("python main.py")
        # os.execl(sys.executable, self.gsdbs.creadPath, *sys.argv)
        # raise SystemExit(0)
        os._exit(0)
        # if self.restartPlanned:
        # else:

    def init(self):
        logging.basicConfig(format='%(asctime)s %(levelname)s:%(message)s', level=self.logginglevel)
        if self.initMethod is not None:
            t1 = threading.Thread(target=self.initMethod, args=[self.gsdbs])
            t1.start()

        l = task.LoopingCall(self.checkThreadRunning)
        l.start(1.0)

        self.reactor.addSystemEventTrigger('before', 'shutdown', self.terminate)

        self.startreactor()

    def startreactor(self):
        if threading.current_thread() is threading.main_thread():
            self.reactor.run()
        else:
            self.reactor.run(installSignalHandlers=0)



import io
import logging
import os

from datetime import datetime, timedelta
import json

import ujson
import yaml
import requests
import ujson as json
from gql import gql, Client
from gql.transport.aiohttp import AIOHTTPTransport
from gql.transport.requests import log as requests_logger, RequestsHTTPTransport
import pandas as pd
import ffmpeg

requests_logger.setLevel(logging.WARNING)


class GSDBS:
    def __init__(self, creadPath, configstr=""):
        self.version = "0.2.6-36"
        self._logger = logging.getLogger(__name__)
        self.credentials = None
        self.accessToken = None
        self.dtablename = None
        self.superdtablename = None
        self.sriBuildInfo = list()
        self.data = pd.DataFrame()
        self.statement = None
        self.creadPath = creadPath
        self.configstr = configstr
        self.readCredentials()
        self.cookiejar = {}
        self.getTokenFromApi()
        self.CGSResponseCode = {}
        self.setupCGSResponseCodes()
        self.debugModus = False
        self.querybuffer = ''
        return

    def setDebugModus(self, debugModus):
        self.debugModus = debugModus

    def readCredentials(self):
        try:
            self._logger.info(f'GS-DBMS Python client version: {self.version}')
            if self.configstr == "":
                f = open(self.creadPath + "/gscred.yml", "r")
                cred = f.read()
                f.close()
                self.credentials = yaml.safe_load(cred)['cred']
            else:
                self.credentials = yaml.safe_load(self.configstr)['cred']
            if (('graphqlapiurl' not in self.credentials.keys())
                    or ('token' not in self.credentials.keys())):
                raise ValueError('error: credentials')
            return
        except Exception as e:
            self._logger.exception(e)
        return

    def getAccessToken(self):
        return self.accessToken["access_token"]

    def getTokenFromApi(self):
        try:
            data = {'username': self.credentials['cnode'],
                    'password': self.credentials['token']}
            self.getAccessTokenFromApi(data)
        except Exception as e:
            self._logger.exception(e)
            os._exit(0)
        return

    def getAccessTokenFromApi(self, data):
        session = requests.session()
        refresh_token_response = session.post(
            self.credentials['baseurl'] + self.credentials['login'],
            json=data,
            headers={'Content-Type': 'application/json'})
        self.accessToken = refresh_token_response.text
        if refresh_token_response.status_code > 299:
            self._logger.warn("Error: Login failed with " + str(refresh_token_response.status_code))
            os._exit(0)

        self.cookiejar = session.cookies
        session.close()

        if 'error' in self.accessToken:
            self._logger.exception("could not authenticate")

    def setupCGSResponseCodes(self):
        self.CGSResponseCode['0'] = 'SUCCESS'
        self.CGSResponseCode['-1'] = 'MISSING (SUPER) DTABLE NAME'
        self.CGSResponseCode['-2'] = 'MISSING DATA LINK LIST'
        self.CGSResponseCode['-3'] = 'MISSING SRI BUILD INFO'
        self.CGSResponseCode['-4'] = 'MISSING DATA'
        self.CGSResponseCode['-5'] = 'MISSING DATALINK FOR SRI BUILD INFO in DATAFRAME'
        self.CGSResponseCode['-6'] = 'QUERY RESULT IS NULL'
        self.CGSResponseCode['-7'] = 'DATAFRAME DATATYPE DOESN\'T MATCHES DTABEL DATATYPE'
        self.CGSResponseCode['-8'] = 'EXECUTION FAILED'
        self.CGSResponseCode['-9'] = 'MISSING STATEMENT TO EXECUTE'
        self.CGSResponseCode['-10'] = 'STATEMENT ERROR'
        self.CGSResponseCode['-11'] = 'ERROR IN DATASCHEMA'
        self.CGSResponseCode['-12'] = 'TYPE ERROR'
        self.CGSResponseCode['-13'] = 'Missing DTABLE'
        self.CGSResponseCode['-14'] = 'Missing DTABLE SCHEMA'
        self.CGSResponseCode['-15'] = 'Missing QUERY'

    def setDTableName(self, dTableName, superDTableName="DTABLE"):
        if dTableName is None or superDTableName is None:
            self.dtablename = None
            self.superdtablename = None
            return -1;
        if (not (isinstance(dTableName, str) and isinstance(superDTableName, str))):
            self.dtablename = None
            self.superdtablename = None
            return -1;
        self.dtablename = dTableName
        self.superdtablename = superDTableName
        return 0

    def clearDTableName(self):
        self.dtablename = None
        self.superdtablename = None
        return 0

    def checkDTableName(self):
        if self.dtablename is None or self.superdtablename is None:
            return -1
        return 0

    def setSriBuildInfo(self, dataLinkList):
        if (dataLinkList is None):
            self.sriBuildInfo = None
            return -2;
        if not isinstance(dataLinkList, list):
            self.sriBuildInfo = None
            return -2
        if (len(dataLinkList) == 0):
            self.sriBuildInfo = None
            return -2
        self.sriBuildInfo = dataLinkList
        return 0

    def getSriBuildInfo(self):
        buidInfo = str()
        for dl in self.sriBuildInfo:
            if len(buidInfo) == 0:
                buidInfo += f'${{{dl}}}'
            else:
                buidInfo += f'-${{{dl}}}'
        return buidInfo

    def clearSriBuildInfo(self):
        self.sriBuildInfo = None
        return 0

    def checkSriBuildInfo(self):
        if (self.sriBuildInfo is None):
            return -3
        if (self.data is None):
            return -4
        for dl in self.sriBuildInfo:
            if dl not in self.data.columns:
                return -5
        return 0

    def setData(self, data):
        if data is None:
            self.data = None
            return -4
        if not isinstance(data, pd.DataFrame):
            self.data == None
            return -4
        if len(data.index) == 0:
            self.data = None
            return -4
        self.data = data
        return 0

    def getGSDBSType(self, pyType):
        if pyType == "object":
            return "STRING"
        elif pyType == "str":
            return "STRING"
        elif pyType == "int":
            return "INTEGER"
        elif pyType == "integer":
            return "INTEGER"
        elif pyType == "int32":
            return "INTEGER"
        elif pyType == "int64":
            return "LONG"
        elif pyType == "float":
            return "FLOAT"
        elif pyType == "float32":
            return "FLOAT"
        elif pyType == "float64":
            return "DOUBLE"
        elif pyType == "bool":
            return "BOOLEAN"
        elif pyType == "datetime":
            return "DATETIME"
        elif pyType == "datetime64":
            return "DATETIME"
        elif pyType == "datetime64[ns]":
            return "DATETIME"
        else:
            return None

    def getDataSchema(self):
        s = str()
        for col in self.data:
            gsDbsType = self.getGSDBSType(str(self.data.dtypes[col]))
            if gsDbsType == None:
                return f"""type error: column {gsDbsType} """
            s += f"""\t\t\t{{alias: "{col}", locale: DE, superPropertyURI: DYNAMIC_DATALINK, DataType: {gsDbsType}}},\n"""
        if len(s) > 0:
            s = s[:-2]
        return s

    def getData(self):
        datalist = list()
        self.data = self.data.astype(str)
        datalist = [self.data.columns.tolist()] + self.flatten(self.data.values.tolist())
        return json.dumps(datalist)

    # s = str()
    # s += f"\t\t\t["
    # for col in self.data:
    #     s += f""""{col}", """
    # s = s[:-2]
    # if len(self.data) > 0:
    #     s += "], \n"
    # else:
    #     s += "]\n"
    #
    # for i in range(len(self.data)):
    #     s += f"\t\t\t["
    #     for col in self.data:
    #
    #         s = str(self.data.loc[i, col])
    #         if s is not None:
    #             if "\"" in s:
    #                 s = s.replace("\"", "'")
    #         s += f""""{s}", """
    #
    #     s = s[:-2]
    #     if i < (len(self.data) - 1):
    #         s += "], \n"
    #     else:
    #         s += "]\n"
    # return s

    def flatten(self, t):
        return [item for item in t]

    def clearData(self):
        self.data = None
        return 0

    def checkData(self):
        if (self.data is None):
            return -4
        for col in self.data:
            gsDbsType = self.getGSDBSType(str(self.data.dtypes[col]))
            if gsDbsType == None:
                return f"""type error: column{gsDbsType} """
        return 0

    def generateMutationStatement(self):
        buildInfo = self.getSriBuildInfo()
        if buildInfo == None:
            return
        dataSchema = self.getDataSchema()
        if dataSchema == None:
            return
        self.statement = \
            f"""mutation {{\n\taddDTable(dtablename: "{self.dtablename}", \n\t\tsuperDTable: [{self.superdtablename}], \n\t\tsriBuildInfo: "{buildInfo}", \n\t\tdataLinks: [\n{dataSchema}\n\t\t], \n\t\tdata: \n{self.getData()}\t\t\n\t)\n}}"""
        return

    async def asyncSchemaCheck(self):
        # check mutation content
        rc = self.checkDTableName()
        if rc != 0: return rc
        rc = self.checkData()
        if rc != 0: return rc
        rc = self.checkSriBuildInfo()
        if rc != 0: return rc

        iQuery = await self.asyncSchemaQuery(self.dtablename)
        if iQuery == None: return -6

        if iQuery['__type'] == None: return 0  # dtable doesn't exist, do what you want
        rc = self.schemaLinkCheck(iQuery)
        return rc

    def schemaCheck(self):
        # check mutation content
        rc = self.checkDTableName()
        if rc != 0: return rc
        rc = self.checkData()
        if rc != 0: return rc
        rc = self.checkSriBuildInfo()
        if rc != 0: return rc

        iQuery = self.schemaQuery(self.dtablename)
        if iQuery == None: return -6

        if iQuery['__type'] == None: return 0  # dtable doesn't exist, do what you want
        rc = self.schemaLinkCheck(iQuery)
        return rc

    def schemaLinkCheck(self, iQuery):
        schema = dict(iQuery['__type'])
        schemaName = schema["name"]
        schemaLinks = schema["fields"]

        for col in self.data:
            dfType = self.getGSDBSType(str(self.data.dtypes[col]))
            linkType = self.getLinkType(schemaLinks, col)
            if dfType.upper() != linkType.upper():
                self._logger.error('ERROR: dataframe datatype ', dfType, ' doesnt\' matches dTable datatype', linkType)
                return -7
        return 0

    def getLinkType(self, schemaLinks, linkName):
        if linkName == None:
            return -1
        for link in schemaLinks:
            if link['name'] == linkName:
                return link['type']['name']
        return None

    async def asyncAddDObject(self, dtablename, sribuildinfo, dataframe, superdtablename="DTABLE", schemaCheck=True):
        self.superdtablename = superdtablename
        rc = self.setDTableName(dtablename)
        if rc != 0: return rc
        rc = self.setSriBuildInfo(sribuildinfo)
        if rc != 0: return rc
        rc = self.setData(dataframe)
        if rc != 0: return rc
        if schemaCheck == True:
            rc = await self.asyncSchemaCheck()
            if rc != 0: return rc
        self.generateMutationStatement()
        if self.debugModus:
            self._logger.debug(self.statement)
        try:
            rc = await self.asyncExecuteStatement()
        except Exception as e:
            self._logger.exception(e)
            return e
        return rc

    def addDObject(self, dtablename, sribuildinfo, dataframe, superdtablename="DTABLE", schemaCheck=True):
        self.superdtablename = superdtablename
        rc = self.setDTableName(dtablename)
        if rc != 0: return rc
        rc = self.setSriBuildInfo(sribuildinfo)
        if rc != 0: return rc
        rc = self.setData(dataframe)
        if rc != 0: return rc
        if schemaCheck == True:
            rc = self.schemaCheck()
            if rc != 0: return rc
        self.generateMutationStatement()
        if self.debugModus:
            self._logger.debug(self.statement)
        try:
            rc = self.executeStatement()
        except Exception as e:
            self._logger.exception(e)
            return e
        return rc

    def addDObjectXSV(self, data):
        self.statement = self.generateStatement(data)
        if self.debugModus:
            self._logger.debug(self.statement)
        try:
            rc = self.executeStatement()
        except Exception as e:
            self._logger.exception(e)
            return e

    def addDObjectv2XSV(self, data):
        self.statement = self.generateV2Statement(data)
        if self.debugModus:
            self._logger.debug(self.statement)
        try:
            rc = self.executeStatement()
        except Exception as e:
            self._logger.exception(e)
            return e

    def updateDObjectXSV(self, data):
        self.statement = self.generateXsvUpdateStatement(data)
        if self.debugModus:
            self._logger.debug(self.statement)
        try:
            rc = self.executeStatement()
        except Exception as e:
            self._logger.exception(e)
            return e

    def deleteDDatalinkXSV(self, data):
        self.statement = self.generateXsvDeleteStatement(data)
        if self.debugModus:
            self._logger.debug(self.statement)
        try:
            rc = self.executeStatement()
        except Exception as e:
            self._logger.exception(e)
            return e

    async def asyncSchemaQuery(self, dtablename):
        statement = "{ __type(name: \"" + dtablename + "\") {name fields { name type { name kind } } } }"

        try:
            result = await self.asyncExecuteStatement(statement)
        except Exception as e:
            self._logger.exception(e)
            return e

        return result

    def schemaQuery(self, dtablename):
        statement = "{ __type(name: \"" + dtablename + "\") {name fields { name type { name kind } } } }"

        try:
            result = self.executeStatement(statement)
        except Exception as e:
            self._logger.exception(e)
            return e

        return result

    async def asyncExecuteStatement(self, statement=None):
        if statement == None:
            if self.statement == None: return -9
            query = gql(self.statement)
        else:
            query = gql(statement)

        try:
            result = await self.prepareClient().execute_async(query)

        except Exception as e:
            os._exit(0)
            if "invalid_token" in str(e):
                self.refreshToken()
                result = await self.asyncExecuteStatement(statement)
            else:
                return e

    def executeStatement(self, statement=None):
        if statement == None:
            if self.statement == None: return -9
            query = gql(self.statement)
        else:
            query = gql(statement)
        try:
            return self.prepareClient().execute(query)
        except Exception as e:
            logging.exception(e)
            os._exit(0)

    def executeStatementSync(self, statement=None):
        if statement == None:
            if self.statement == None: return -9
            query = gql(self.statement)
        else:
            query = gql(statement)
        try:
            return self.prepareClientSync().execute(query)
        except Exception as e:
            logging.exception(e)
            os._exit(0)

    def prepareClient(self):
        return Client(
            transport=AIOHTTPTransport(url=self.credentials["baseurl"] + self.credentials['graphqlapiurl'],
                                       cookies=self.cookiejar), execute_timeout=40)

    def prepareClientSync(self):
        return Client(
            transport=RequestsHTTPTransport(url=self.credentials["baseurl"] + self.credentials['graphqlapiurl'],
                                            cookies=self.cookiejar), execute_timeout=40)

    async def asyncDropDTable(self, dtablename):
        statement = f"""mutation {{\n\tdropDTable(dtablename: {dtablename})\n}}"""
        result = await self.asyncExecuteStatement(statement)
        return result

    def dropDTable(self, dtablename):
        statement = f"""mutation {{\n\tdropDTable(dtablename: {dtablename})\n}}"""
        result = self.executeStatement(statement)
        return result

    def getDTableFullQuery(self, dtablename):
        try:
            self.querybuffer = 'query {\n'
            self.querybuffer += dtablename + '\n'
            self.getDTableQuery(dtablename)
            self.querybuffer += '}'
        except Exception as e:
            return e
        return self.querybuffer

    def getDTableQuery(self, dtablename):
        statement = f"""{{
              __type(name: \"{dtablename}\") {{
                name
                fields {{
                  name
                  type {{
                    name
                    kind
                    ofType {{
                      name
                      kind
                    }}
                  }}
                }}
              }}
            }}"""
        result = self.executeStatement(statement)
        if result == None:
            raise ValueError('MISSING DTABLE')

        schema = dict(result['__type'])
        if schema == None:
            raise ValueError('MISSING DTABLE SCHEMA')

        schemaName = schema["name"]

        self.querybuffer += '{\n'
        for field in schema["fields"]:
            if field['name'] == 'episodes':
                continue
            if field['name'] == 'dynamic_snapshot_timestamp':
                continue
            if field['name'] == 'totalrowcount':
                continue
            self.querybuffer += field['name'] + "\n"
            if field['type'] != None:
                if field['type']['ofType'] != None:
                    self.getDTableQuery(field['type']['ofType']['name'])
        self.querybuffer += '}\n'
        return 0

    def getForeignDTableName(self, dtablename, foreignLinkName):
        statement = f"""{{
              __type(name: \"{dtablename}\") {{
                name
                fields {{
                  name
                  type {{
                    name
                    kind
                    ofType {{
                      name
                      kind
                    }}
                  }}
                }}
              }}
            }}"""
        result = self.executeStatement(statement)
        if result == None:
            raise ValueError('MISSING DTABLE')
        schema = dict(result['__type'])
        if schema == None:
            raise ValueError('MISSING DTABLE SCHEMA')
        schemaName = schema["name"]

        for field in schema["fields"]:
            if field['name'] == foreignLinkName:
                if field['type'] != None:
                    if field['type']['ofType'] != None:
                        return field['type']['ofType']['name']
        return None

    def getFileFromCloud(self, filename):
        filename = filename.replace("/", "$")
        storageURL = self.credentials["baseurl"] + self.credentials['storageurl'] + "/file/" + filename
        # headers = {'Authorization': 'Bearer ' + self.accessToken['access_token']}
        resp = requests.get(storageURL, cookies=self.cookiejar)
        buf = io.BytesIO(resp.content)
        return buf

    def saveFileInCloud(self, filenamewithpath, fileasbytesio, fileformat):
        storageURL = self.credentials["baseurl"] + self.credentials['storageurl'] + "/file"
        # headers = {'Authorization': 'Bearer ' + self.accessToken['access_token']}
        files = [
            ('filePart', (filenamewithpath, fileasbytesio.getvalue(), fileformat)),
        ]
        res = requests.post(storageURL, cookies=self.cookiejar, files=files)

    def generateStatement(self, data):
        nl = '"'
        return f"""mutation 
        {{  addDTable(
                dtablename: "{data["dtablename"]}",
                superDTable: [{data["superdtable"].upper()}],
                sriBuildInfo: "{"-".join(map(lambda x: f"${{{x}}}", data["sribuildinfo"]))}",
                dataLinks: [{",".join(
            map(lambda x:
                f'{{alias:"{x["name"]}", locale: DE, superPropertyURI: DYNAMIC_DATALINK, DataType: {x["type"].upper()}}}'
                , data["header"]))}],
                data:
                [ [{",".join(map(lambda x: f'"{x["name"]}"', data["header"]))}]
                  {",".join(
            map(
                lambda row: f'[{",".join(map(lambda col: f"{nl}{col}{nl}", row))}]', data["rowdata"]))}
                ]
                )}}"""

    def generateV2Statement(self, data):
        nl = '"'
        empty = ""
        ignore = "IGNORE"
        dontignore = "DONTIGNORE"

        for idx, rowdata in enumerate(data["rowdata"]):
            for idx2, coldata in enumerate(rowdata):
                if isinstance(coldata, str):
                    data["rowdata"][idx][idx2] = ujson.dumps(coldata)
                else:
                    data["rowdata"][idx][idx2] = f'"{coldata}"'

        return f"""mutation 
        {{  addDTablev2(
                dtablename: "{data["dtablename"]}",
                superDTable: [{data["superdtable"].upper()}],
                sriBuildInfo: "{"-".join(map(lambda x: f"${{{x}}}", data["sribuildinfo"]))}",
                dataLinks: [{",".join(
            map(lambda x:
                f'{{alias:"{x["name"]}", locale: DE, superPropertyURI: DYNAMIC_DATALINK, DataType: {x["type"].upper()}}}'
                , data["header"]))}],
                data:
                [ [{",".join(map(lambda x: f'{{value:{nl}{x["name"]}{nl}}}', data["header"]))}]
                  {",".join(
            map(
                lambda row: f'[{",".join(map(lambda col: f"{{value:{empty if col is None else col},state:{ignore if col is None else dontignore}}}", row))}]', data["rowdata"]))}
                ]
                )}}"""

    def generateXsvUpdateStatement(self, data):
        stmnt = f'mutation{{' \
                f'updateDTable(dtablename:{data["dtablename"]}' \
                f' where: ['
        for e in data['where']:
            connective = e['connective']
            column = e['column']
            operator = e['operator']
            value = e['value']
            stmnt += f'{{connective: {connective}, column: {column}, operator: {operator}, value: "{value}"}}'
        stmnt += f']' \
                 f' updatelist: ['
        for key, value in data['update'].items():
            stmnt += f'{{datalink: {key}, value: "{value}"}}'
        stmnt += f'])}}'
        return stmnt

    def generateXsvDeleteStatement(self, data):
        stmnt = f'mutation{{' \
                f'deleteDDatalink(dtablename:{data["dtablename"]}' \
                f' where: ['
        for e in data['where']:
            connective = e['connective']
            column = e['column']
            operator = e['operator']
            value = e['value']
            stmnt += f'{{connective: {connective}, column: {column}, operator: {operator}, value: "{value}"}}'
        stmnt += f']' \
                 f' updatelist: ['
        for key, value in data['update'].items():
            stmnt += f'{{datalink: {key}}}'
        stmnt += f'])}}'
        return stmnt

    def saveVod(self, target, vodpath, datetime):

        # vodFileName = vodpath.rsplit("\\", 1)[1]

        head, vodFileName = os.path.split(vodpath)

        duration = int(float(ffmpeg.probe(vodpath)["format"]["duration"]))

        storageURL = self.credentials["baseurl"] + self.credentials['storageurl'] + f"/vod/{target}"
        with open(vodpath, 'rb') as fh:
            buf = io.BytesIO(fh.read())
            files = [
                ('filePart', (f"{vodFileName}", buf.getvalue(), "application/octet-stream")),
            ]
            res = requests.post(storageURL, cookies=self.cookiejar, files=files)

        self.executeStatement(f"""
                mutation{{
                addDTable(
                    dtablename:"gsvod",
                    superDTable:[DTABLE],
                    sriBuildInfo:"${{streamkey}}-${{fragmentid}}",
                    dataLinks:[{{ alias:"streamkey",locale:DE,superPropertyURI:DYNAMIC_DATALINK,DataType:STRING}},
                    {{ alias:"fragmentid",locale:DE,superPropertyURI:DYNAMIC_DATALINK,DataType:STRING}},
                    {{ alias:"starttime",locale:DE,superPropertyURI:DYNAMIC_DATALINK,DataType:DATETIME}},
                    {{ alias:"segmentlength",locale:DE,superPropertyURI:DYNAMIC_DATALINK,DataType:STRING}}],
                    data:[["streamkey","fragmentid","starttime","segmentlength"],
                          ["{target}","{vodFileName}","{datetime}","{duration}"]
                          ])
            }}
            """)
        self._logger.debug("saved vod:" + vodpath)